"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.

Extension types defined here:
2: contains raw bytes, interpreted as unsigned bignum
3: Path, as a msgpack object stream of its elements
4: contains raw bytes, interpreted as UTF-8, returned as (named) Proxy object
"""

from functools import partial

import msgpack

from ._dict import attrdict
from ._path import Path

__all__ = ["packer", "unpacker", "stream_unpacker", "Proxy"]


class Proxy:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name !r})"


def _encode(data):
    if isinstance(data, int) and data >= 1 << 64:
        return msgpack.ExtType(2, data.to_bytes((data.bit_length() + 7) // 8, "big"))
    elif isinstance(data, Path):
        return msgpack.ExtType(3, b"".join(packer(x) for x in data))
    elif isinstance(data, Proxy):
        return msgpack.ExtType(4, data.name.encode("utf-8"))
    return data


def _decode(code, data):
    if code == 2:
        return int.from_bytes(data, "big")
    elif code == 3:
        s = stream_unpacker()
        s.feed(data)
        return Path(*s)
    elif code == 4:
        try:
            return Proxy(data.decode("utf-8"))
        except UnicodeDecodeError:
            return Proxy(str(data))
    return msgpack.ExtType(code, data)


# single message packer
packer = partial(
    msgpack.packb,
    strict_types=False,
    use_bin_type=True,
    default=_encode
)

# single message unpacker
unpacker = partial(
    msgpack.unpackb,
    object_pairs_hook=attrdict,
    strict_map_key=False,
    raw=False,
    use_list=False,
    ext_hook=_decode,
)

# stream unpacker factory
stream_unpacker = partial(
    msgpack.Unpacker,
    object_pairs_hook=attrdict,
    strict_map_key=False,
    raw=False,
    use_list=False,
    ext_hook=_decode,
)



# Some rudimentary tests for merge and combine_dict

from distkv.util import NotGiven, combine_dict, merge


def chkm(a, b, c, drop=False):
    r = merge(a, b, drop=drop)
    assert r == c


def chkc(a, b, c):
    r = combine_dict(a, b)
    assert r == c


def test_merge():
    chkm(dict(a=1, b=2, c=3), dict(b=NotGiven), dict(a=1, c=3))
    chkm(dict(a=1, b=2, c=3), dict(b=4, d=5), dict(a=1, b=4, c=3, d=5))
    chkm(dict(a=1, b=[1, 2, 3], c=3), dict(b=(4, NotGiven, None, 6)), dict(a=1, b=[4, 3, 6], c=3))
    chkm(
        dict(a=1, b=[1, 2, 3], c=3), dict(b={0: 4, 1: NotGiven, 3: 6}), dict(a=1, b=[4, 3, 6], c=3)
    )
    chkm(
        dict(a=1, b=[1, 2, 3], c=3),
        dict(a=1, b=(4, NotGiven, None, 6)),
        dict(a=1, b=[4, 3, 6]),
        drop=True,
    )


def test_combine():
    chkc(dict(a=1, b=2, c=3), dict(b=4, d=5), dict(a=1, b=2, c=3, d=5))
    chkc(dict(a=1, b=2, c=3), dict(b=NotGiven), dict(a=1, b=2, c=3))
    chkc(dict(b=NotGiven), dict(a=1, b=2, c=3), dict(a=1, c=3))



__all__ = ["merge"]


from . import NotGiven


def _merge_dict(d, other, drop=False):
    for key, value in other.items():
        if value is NotGiven:
            d.pop(key, None)
        elif key in d:
            d[key] = _merge_one(d[key], value, drop=drop)
        else:
            d[key] = value

    if drop:
        keys = []
        for k in d.keys():
            if k not in other:
                keys.append(k)
        for k in keys:
            del d[k]


def _merge_list(item, value, drop=False):
    off = 0
    if isinstance(value, (list, tuple)):
        # two lists
        lim = len(item)
        for i in range(min(lim, len(value))):
            if value[i] is NotGiven:
                item.pop(i - off)
                off += 1
            else:
                item[i - off] = _merge_one(item[i - off], value[i], drop=drop)

        while len(item) + off < len(value):
            val = value[len(item) + off]
            if val is NotGiven:
                off += 1
            else:
                item.append(val)

        if drop:
            while len(item) + off > len(value):
                item.pop()

    else:
        # a list, and a dict with updated/new/deleted values
        for i in range(len(item)):
            if i in value:
                val = value[i]
                if val is NotGiven:
                    item.pop(i - off)
                    off += 1
                else:
                    item[i - off] = _merge_one(item[i - off], val, drop=drop)

        while len(item) + off in value:
            val = value[len(item) + off]
            if val is NotGiven:
                off += 1
            else:
                item.append(val)


def _merge_one(d, other, drop=False):
    if isinstance(d, dict):
        if isinstance(other, dict):
            _merge_dict(d, other, drop=drop)
        else:
            return other
    elif isinstance(d, list):
        if isinstance(other, (dict, list, tuple)):
            _merge_list(d, other, drop=drop)
        else:
            return other
    else:
        return d if other is None else other
    return d


def merge(d, *others, drop=False):
    """
    Deep-merge two data structures. Values from later structures overwrite earlier ones.
    Values of `NotGiven` are deleted.

    Rules:
    * Two dicts recurses into same-key items, and adds new items
    * Two lists recurses into same-index items, and appends from a longer list
    * A list and a dict treats the dict as a sparse list. The value of numeric keys just
      beyond the list's length are appended, others are ignored.
    * otherwise, returns the second argument if it is not None, otherwise the first

    This applies recursively.

    If "drop" is given, delete source keys that are not in the destination.
    This is useful for in-place replacements.
    """
    for other in others:
        d = _merge_one(d, other, drop=drop)
    return d
